/*******************************************************************************
    Copyright (c) 2013 NVidia Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/

#include "uvm_ioctl.h"
#include "nvidia_uvm_common.h"
#include "nvidia_uvm_lite.h"
#include "nvidia_uvm_lite_counters.h"

#if defined (DEBUG)
#include "uvm_gpu_ops_tests.h"
#endif

//
// nvidia_uvm_lite_api.c
//
// This file contains code for UVM API calls that are issued via ioctl()
// call.
//

// Locking: needs mmap_sem for read
static struct vm_area_struct *
find_uvmlite_vma(unsigned long long requestedBase,
    unsigned long long length,
    struct file *filp)
{
    struct vm_area_struct *vma = find_vma(current->mm, requestedBase);
    if (vma == NULL)
        return NULL;
    if (vma->vm_file != filp)
        return NULL;
    if (vma->vm_start != requestedBase || vma->vm_end != PAGE_ALIGN(requestedBase + length))
        return NULL;
    return vma;
}

UVM_STATUS
uvmAPI_ReserveVa(UVM_RESERVE_VA_PARAMS *pParams, struct file *filp)
{
    pr_debug_ratelimited("%s: requestedBase: 0x%llx, length: 0x%llx\n",
                __FUNCTION__, pParams->requestedBase, pParams->length);
    //
    // There is nothing required here yet. The mmap() call in user space handles
    // everything.
    //
    return UVM_OK;
}

UVM_STATUS
uvmAPI_ReleaseVa(UVM_RELEASE_VA_PARAMS *pParams, struct file *filp)
{
    pr_debug_ratelimited("%s: requestedBase: 0x%llx, length: 0x%llx\n",
             __FUNCTION__, pParams->requestedBase, pParams->length);
    //
    // There is nothing required here yet. The mmap() call in user space handles
    // everything.
    //
    return UVM_OK;
}

UVM_STATUS
uvmAPI_RegionCommit(UVM_REGION_COMMIT_PARAMS *pParams, struct file *filp)
{
    UvmCommitRecord * pRecord;
    DriverPrivate* pPriv = (DriverPrivate*)filp->private_data;
    struct vm_area_struct * vma;
    UVM_STATUS uvmStatus = UVM_OK;

    pr_debug_ratelimited("%s: requestedBase: 0x%llx, length: 0x%llx, "
                         "streamId: 0x%llx\n",
             __FUNCTION__, pParams->requestedBase, pParams->length,
             pParams->streamId);

    down_write(&current->mm->mmap_sem);

    vma = find_uvmlite_vma(pParams->requestedBase, pParams->length, filp);

    // Make sure the vma belongs to the driver by checking the vm_file
    if (vma == NULL)
    {
        up_write(&current->mm->mmap_sem);
        return UVM_ERR_ADDRESS_IN_USE;
    }

    //
    // TODO: is it better to simply do all of the actions for UvmRegionCommit,
    // in the uvmlite_mmap() callback, instead of doing part in the mmap, and
    // this last bit here?
    //
    // Putting everything in uvmlite_mmap would make UvmRegionCommit look mostly
    // atomic, from user space. And there is no real need or benefit to
    // splitting it up between and mmap call, and an ioctl call.
    //
    // The ioctl needs to at least assign the stream so some splitting of work
    // is required.
    //
    down_write(&pPriv->uvmPrivLock);
    pRecord = uvmlite_create_commit_record(pParams, pPriv, vma);
    if (NULL == pRecord)
        uvmStatus = UVM_ERR_FAILED; // TODO: pick a better return code

    up_write(&pPriv->uvmPrivLock);

    vma->vm_private_data  = pRecord;
    up_write(&current->mm->mmap_sem);

    return uvmStatus;
}

UVM_STATUS
uvmAPI_RegionDecommit(UVM_REGION_DECOMMIT_PARAMS *pParams, struct file *filp)
{
    pr_debug_ratelimited("%s: requestedBase: 0x%llx, length: 0x%llx\n",
             __FUNCTION__, pParams->requestedBase, pParams->length);
    //
    // There is nothing required here yet. The vma.close callback handles
    // everything.
    //
    return UVM_OK;
}

UVM_STATUS
uvmAPI_RegionSetStream(UVM_REGION_SET_STREAM_PARAMS *pParams,
                       struct file *filp)
{
    DriverPrivate* pPriv = (DriverPrivate*)filp->private_data;
    UVM_STATUS uvmStatus;
    struct vm_area_struct * vma;
    UvmCommitRecord * pRecord;

    pr_debug_ratelimited("%s: requestedBase: 0x%llx, length: 0x%llx, "
             "newStreamId: 0x%llx\n",
             __FUNCTION__, pParams->requestedBase, pParams->length,
             pParams->newStreamId);

    down_write(&current->mm->mmap_sem);

    vma = find_uvmlite_vma(pParams->requestedBase, pParams->length, filp);
    if (vma == NULL)
    {
        up_write(&current->mm->mmap_sem);
        return UVM_ERR_ADDRESS_IN_USE;
    }

    down_write(&pPriv->uvmPrivLock);
    pRecord = (UvmCommitRecord*)vma->vm_private_data;

    uvmStatus = uvmlite_region_set_stream(pRecord, pParams->newStreamId);

    up_write(&pPriv->uvmPrivLock);
    up_write(&current->mm->mmap_sem);

    return uvmStatus;
}

UVM_STATUS
uvmAPI_SetStreamRunning(UVM_SET_STREAM_RUNNING_PARAMS *pParams,
                        struct file *filp)
{
    DriverPrivate* pPriv = (DriverPrivate*)filp->private_data;
    UVM_STATUS uvmStatus;

    pr_debug_ratelimited("%s: streamID: 0x%llx\n", __FUNCTION__,
                         pParams->streamId);

    down_write(&current->mm->mmap_sem);
    down_write(&pPriv->uvmPrivLock);
    uvmStatus = uvmlite_set_stream_running(pPriv, pParams->streamId);
    up_write(&pPriv->uvmPrivLock);
    up_write(&current->mm->mmap_sem);

    return uvmStatus;
}

UVM_STATUS
uvmAPI_SetStreamStopped(UVM_SET_STREAM_STOPPED_PARAMS *pParams,
                        struct file *filp)
{
    DriverPrivate* pPriv = (DriverPrivate*)filp->private_data;
    UvmSize lastIndex = 0;
    UVM_STATUS uvmStatus;

    if (pParams->nStreams > 1)
    {
        lastIndex = pParams->nStreams - 1;
    }

    pr_debug_ratelimited("%s: streamIDs 0x%llx - 0x%llx\n",
             __FUNCTION__, pParams->streamIdArray[0],
             pParams->streamIdArray[lastIndex]);

    down_write(&current->mm->mmap_sem);
    down_write(&pPriv->uvmPrivLock);
    uvmStatus = uvmlite_set_streams_stopped(pPriv, pParams->streamIdArray,
                                           pParams->nStreams);
    up_write(&pPriv->uvmPrivLock);
    up_write(&current->mm->mmap_sem);
    return uvmStatus;
}

UVM_STATUS
uvmAPI_MigrateToGpu(UVM_MIGRATE_TO_GPU_PARAMS *pParams, struct file *filp)
{
    DriverPrivate* pPriv = (DriverPrivate*)filp->private_data;
    UVM_STATUS uvmStatus;
    struct vm_area_struct * vma;
    UvmCommitRecord * pRecord;

    pr_debug_ratelimited("%s: requestedBase: 0x%llx, length: 0x%llx, "
             "flags: 0x%x\n",
             __FUNCTION__, pParams->requestedBase, pParams->length,
             pParams->flags);

    down_write(&current->mm->mmap_sem);

    vma = find_uvmlite_vma(pParams->requestedBase, pParams->length, filp);
    if (vma == NULL)
    {
        up_write(&current->mm->mmap_sem);
        return UVM_ERR_ADDRESS_IN_USE;
    }

    down_write(&pPriv->uvmPrivLock);
    pRecord = (UvmCommitRecord*)vma->vm_private_data;

    uvmStatus = uvmlite_migrate_to_gpu(pParams->requestedBase,
                                       pParams->length,
                                       pParams->flags,
                                       vma,
                                       pRecord);
    up_write(&pPriv->uvmPrivLock);
    up_write(&current->mm->mmap_sem);

    return uvmStatus;
}

#if defined(DEBUG)
UVM_STATUS
uvmAPI_RunTest(UVM_RUN_TEST_PARAMS *pParams, struct file *filp)
{
    switch (pParams->testNumber)
    {
        case UVM_GPU_OPS_SAMPLE_TEST:
            gpuOpsSampleTest(&pParams->gpuUuid);
            break;
        case UVM_PAGE_MIGRATION_TEST:
            pageMigrationTest(&pParams->gpuUuid);
            uvm_print_uuid(__FUNCTION__, "Entering", &pParams->gpuUuid);
            break;
        default:
            pr_info("%s: bad test number: 0x%x\n",
                     __FUNCTION__, pParams->testNumber);
            return UVM_ERR_INVALID_ARGUMENT;
    }

    return UVM_OK;
}
#endif

UVM_STATUS
uvmAPI_AddSession(UVM_ADD_SESSION_PARAMS *pParams, struct file *filp)
{
    UVM_STATUS uvmStatus;
    DriverPrivate* pPriv = (DriverPrivate*)filp->private_data;
    UvmProcessCounterInfo *pProcessCounterInfo = NULL;
    uid_t euid = UVM_ROOT_UID;
    //
    // TODO Bug: 1372634 validate countersBaseAddress. In particular,
    // ensure that the owning vma has read-only protections.
    //
    uvmStatus = uvmlite_secure_get_process_counter_info(pParams->pidTarget,
                                                        &pProcessCounterInfo,
                                                        &euid);
    if (uvmStatus != UVM_OK)
        return uvmStatus;

    uvmStatus = NvUvmAddSessionInfo(pParams->pidTarget, euid, 
                                    &pParams->sessionIndex,
                                    pProcessCounterInfo, 
                                    pParams->countersBaseAddress,
                                    &pPriv->processRecord);

    if (uvmStatus != UVM_OK)
        uvmlite_put_and_unrefcount_process_counter_info(pProcessCounterInfo);
    
    return uvmStatus;
}

UVM_STATUS
uvmAPI_RemoveSession(UVM_REMOVE_SESSION_PARAMS *pParams, struct file *filp)
{
    UVM_STATUS uvmStatus;
    DriverPrivate* pPriv = (DriverPrivate*)filp->private_data;
    uvmStatus = NvUvmRemoveSessionInfo(pParams->sessionIndex,
                                        &pPriv->processRecord);
                                       
    return uvmStatus;
}

UVM_STATUS
uvmAPI_EnableCounters(UVM_ENABLE_COUNTERS_PARAMS *pParams, struct file *filp)
{
    UVM_STATUS uvmStatus;
    DriverPrivate* pPriv = (DriverPrivate*)filp->private_data;
    uvmStatus = NvUvmCounterStateUpdate(pParams->sessionIndex,
                                        pParams->config,
                                        pParams->count,
                                        &pPriv->processRecord);
    return uvmStatus;
}

UVM_STATUS
uvmAPI_MapCounter(UVM_MAP_COUNTER_PARAMS *pParams, struct file *filp)
{
    UVM_STATUS uvmStatus;
    DriverPrivate* pPriv = (DriverPrivate*)filp->private_data;
    uvmStatus = NvUvmMapCounter(pParams->sessionIndex,
                                pParams->scope,
                                pParams->counterName,
                                pParams->gpuUuid,
                                &pPriv->processRecord,
                                (void **)&pParams->addr);
    return uvmStatus;
}
